/***************************************************************************
 *   Copyright (C) 2007 by E. Angelou, A. Sottoriva                        *
 *   Universiteit van Amsterdam                                            *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#ifndef BASIC_H_
#define BASIC_H_

#define SIZE_PARMS 0
#define DEBUG 0

#include <iostream>
#include <fstream>
#include <cstdlib>
#include <fstream>
#include <string>
#include <list>
#include <vector>
#include <iterator>
#include <math.h>
// #include <mpi/mpi.h>
#include <mpi.h>

using namespace std;

int mubins;			// total number of theta bins (0, 90)
int nlevel;			// total number of levels in the slab
int phibins;			// total number of phi bins (0, 360)
int seed;			// random generator seed
int nphotons; 		// total number of photons

double* energy;		// energy per bin
double* erri;			// error per bin
double* intensity;		// intensity per bin
double* sigmai;		// cross action per bin
double* theta;		// theta of a bin

double density;		// number of mass of the slab
double opacity;		// cross section of the slab
double distanc;              // measure proportional to the distance from the observer
double depth;			// slab depth
double fi, fq, fu, fv;	// Stoke's parameters
double g2;			// hdd^2
double pl;			// peak linear polarization
double pc;			// peak circular polarization
double sc;			// skewing factor
double hgg;			// scattering asimmetry parameter
double obs_theta;		// observer's theta
double obs_phi;		// observer's phi
double obs_sint;
double obs_cost;
double obs_sinp;
double obs_cosp;
int xdim,ydim,zdim; //x,y,z dimention of the grid
int xcell,ycell,zcell; //current cell in the grid
double xmax,ymax,zmax; //grid span


// for each level of the slab we compute the moments:
// J = mean intensity
// H = mean flux
// K = mean radiation pressure
double* jplus;
double* jminus;
double* hplus;
double* hminus;
double* kplus;
double* kminus;

//The arrays of the cell faces
double* xfaces;
double* yfaces;
double* zfaces;

//The array of the cell coordinates
// double**** coords;

//The array of the cell density
double*** rhokappa;


int image_size;
int* image;


// find maximum tau the photon can travel


// initialization of the arrays
void init_array()
{
	double dthet, halfw, dphi, halfh;

	energy = new double[mubins];
	intensity = new double[mubins];
	erri = new double[mubins];
	sigmai = new double[mubins];
	theta = new double[mubins];

	xfaces= new double [xdim];
	yfaces= new double [ydim];
	zfaces= new double [zdim];

	for (int i = 0; i < mubins; ++i)
	{
		energy[i] = 0.0;
		sigmai[i] = 0.0;
		erri[i] = 0.0;
		intensity[i] = 0.0;
	}

	jplus = new double[nlevel];

	jminus = new double[nlevel];
	hplus = new double[nlevel];
	hminus = new double[nlevel];
	kplus = new double[nlevel];
	kminus = new double[nlevel];

	for (int j = 0; j < nlevel; ++j)
	{
		jplus[j] = 0.0;
		jminus[j] = 0.0;
		hplus[j] = 0.0;
		hminus[j] = 0.0;
		kplus[j] = 0.0;
		kminus[j] = 0.0;
	}

	image = new int[image_size * image_size];

	memset(image, 0, sizeof(int) * image_size * image_size);

	dthet = 2.0 / mubins;	// chop theta in mubins bins
	halfw = dthet / 2.0;	// half of a theta bin

	int rj;

	// assign an angle for each of the theta bins
	for (int j = 1; j <= mubins; ++j)
	{
		rj = j - 1;
		theta[j - 1] = acos(rj * dthet + halfw - 1.0) * 180.0 / M_PI - 90.0;
	}

	obs_sint = sin(obs_theta);

	obs_cost = cos(obs_theta);
	obs_sinp = sin(obs_phi);
	obs_cosp = cos(obs_phi);

	//setup face coords

	for (int i=0; i<xdim; i++)
	{
		xfaces[i]=i*2.0*xmax/xdim;
	}

	for (int i=0; i<ydim; i++)
	{
		yfaces[i]=i*2.0*ymax/ydim;
	}

	for (int i=0; i<zdim; i++)
	{
		zfaces[i]=i*2.0*zmax/zdim;
	}


	//Create the density array
	rhokappa = new double**[xdim];

	for (int i=0; i<xdim; i++)
		rhokappa[i] = new double*[ydim];

	for (int i=0; i<xdim; i++)
		for (int j=0; j<ydim; j++)
			rhokappa[i][j]=new double[zdim];

	//initialize the density array as needed for each experiment, change from here on to create fancy dust clouds :-)
	for (int i=0; i<xdim; i++)
	{
		for (int j=0; j<ydim; j++)
		{
			for (int k=0; k<zdim; k++)
			{

				// 				//initialize density grid
				// 				if (j>7&&j<13){
				rhokappa[i][j][k] = density * opacity;
// 		rhokappa[i][j][k] = density * opacity * 1.5e13;
				// 				}
				// 				else
				// 				{
				// 					rhokappa[i][j][k] = 0.1;
				// 				}

				// 				cout << "RHOKAPPA:" << i <<" "<<j <<" "<<k <<" "<< rhokappa[i][j][k] <<endl;
			}
		}
	}


}

// print output
void output(string fname)
{
	double tot_energy = 0;
	double tot_intensity = 0;

	fstream file_op1(fname.c_str(),ios::out);

	// compute intensities and energies for each bins
	file_op1<<"\nTHETA\t\t ENERGY\t\t SIGMA\t\t INTENSITY"<<endl;
	file_op1<<"|================================================================|"<<endl;

	for (int j = 0; j < mubins; ++j)
	{
		intensity[j] = energy[j] / (2.0 * nphotons * cos(theta[j] * M_PI/180.0)) * mubins;
		sigmai[j] = sqrt(erri[j]) / nphotons;
		energy[j] = energy[j] / nphotons;
		tot_energy += energy[j];
		tot_intensity += intensity[j];
		file_op1 <<theta[j] <<"\t\t" <<  energy[j]<< "\t\t" << sigmai[j] << "\t\t" << intensity[j] << endl;
	}

	file_op1<<"|================================================================|"<<endl;

	file_op1 << "Total energy: " << tot_energy << " \t\tTotal intensity: " << tot_intensity <<endl;
	file_op1<<"|================================================================|"<<endl;
	file_op1<<"|================================================================|"<<endl;
	file_op1<<"\n\n\n"<<endl;
	file_op1 <<"\njplus\t\t jminus\t\t hplus\t\t hminus\t\t kplus\t\t kminus"<<endl;
	file_op1 <<"|============================================================================================|"<<endl;
	// update moments arrays

	for (int j = 0; j < nlevel; j++)
	{
		jplus[j] = jplus[j] / nphotons;
		jminus[j] = jminus[j] / nphotons;
		hplus[j] = hplus[j] / nphotons;
		hminus[j] = hminus[j] / nphotons;
		kplus[j] = kplus[j] / nphotons;
		kminus[j] = kminus[j] / nphotons;
		file_op1 << jplus[j] <<"\t\t"<<jminus[j]<<"\t\t"<<hplus[j]<<"\t\t"<<hminus[j]<<"\t\t"<<kplus[j]<<"\t\t"<<kminus[j]<<endl;
	}

	file_op1 <<"|============================================================================================|"<<endl;


	file_op1.close();
}

// generate a rundom number uniformly distributed in (0, 1)
double urand()
{
	return (double)random() / RAND_MAX;
}

// emit photons isotropically from the origin
void newphoton(double& sint, double& cost, double& sinp, double& cosp, double& theta, double& phi, double& xtot, double& ytot, double& ztot, double& nxp, double& nyp, double& nzp, int& xcell, int& ycell, int& zcell)
{
	theta = (M_PI * urand()) - M_PI/2;
	cost = cos(theta);
	sint = sin(theta);
	phi = 2.0 * M_PI * urand();
	cosp = cos(phi);
	sinp = sin(phi);
	xtot = 0.0;
	ytot = 0.0;
	ztot = 0.0;
	fi = 1.0;
	fq = fu = fv = 0.0;

	nxp = sint * cosp;
	nyp = sint * sinp;
	nzp = cost;

	//     xcell = (int) (xdim * (xtot + xmax) / (xmax * 2.0))-1;
	//     ycell = (int) (ydim * (ytot + ymax) / (ymax * 2.0))-1;
	//     zcell = (int) (zdim * (ztot + zmax) / (zmax * 2.0))-1;
	xcell =0;
	ycell =0;
	zcell =0;
}

// isotropical scatterin
void isoscatt(double& sint, double& cost, double& sinp, double& cosp, double& theta, double& phi, double& nxp,double& nyp,double& nzp)
{
	//     cost = 2.0 * urand() - 1.;
	//     sint = sqrt(1.0 - cost * cost);
	theta = (M_PI * urand()) - M_PI/2; // M_PI * urand();
	cost = cos(theta);
	sint = sin(theta);
	phi = 2.0 * M_PI * urand();
	cosp = cos(phi);
	sinp = sin(phi);
	nxp = sint * cosp;
	nyp = sint * sinp;
	nzp = cost;

}



// compute the moments at nlevel+1 levels in a plane parallel slab (the moments are split into + and - components
// that contribute to the final J, H and K moments.
void moments(double& x1, double& y1, double& z1, double& x2, double& y2, double& z2, double& cost, int nlevel)
{
	int i, l1, l2;

	if ((z1 > 0.0) && (z2 > 0.0) && ((int)z1 * nlevel == (int)z2 * nlevel))
	{
		return;
	}

	if (cost > 0.0)
	{
		if (z1 <= 0.0)
		{
			l1 = 1;
		}
		else
		{
			l1 = int(z1 * nlevel) + 2;
		}

		if (z2 >= 1.0)
		{
			l2 = nlevel + 1;
		}
		else
		{
			l2 = int(z2* nlevel) + 1;
		}

		for (i = l1; i <= l2; i++)
		{
			jplus[i-1] = jplus[i-1] + 1.0 / cost;
			hplus[i-1] = hplus[i-1] + 1.0;
			kplus[i-1] = kplus[i-1] + cost;
		}
	}
	else
	{
		
		if (cost < 0.0)
		{
			l1 = (int)z1 * nlevel + 1;

			if (z2 <= 0.0)
			{
				l2 = 1;
			}
			else
			{
				l2 = (int)z2 * nlevel + 2;
			}

			for (i = l2; i <= l1; i++)
			{
				jminus[i-1] = jminus[i-1] + 1.0 / (abs((int)cost));
				hminus[i-1] = hminus[i-1] - 1;
				kminus[i-1] = kminus[i-1] + abs((int)cost);
			}
		}
	}

}

//Abort program
void abortMPI(int err, int rank)
{
	char error_str[MPI_MAX_ERROR_STRING];
	int len;
	int init_flag;

	MPI_Error_string(err, error_str, &len);

	if (rank < 0)
	{
		fprintf(stderr, "Error from %d: %s (errno = %d)\n", rank, error_str, err);
	}
	else
	{
		fprintf(stderr, "Global error: %s (errno = %d)\n", error_str, err);
	}

	MPI_Initialized(&init_flag);

	if (init_flag)
		MPI_Abort(MPI_COMM_WORLD, EXIT_FAILURE);

	exit(EXIT_FAILURE);
}

#endif
