/***************************************************************************
 *   Copyright (C) 2007 by E. Angelou, A. Sottoriva                        *
 *   Universiteit van Amsterdam                                            *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
// #include <mpi.h>
// ../../bin/date; mpirun -np 13 /home/eangelou/mpirad2/debug/src/mpirad2 /home/eangelou/mpirad2/src/slab.par
#include "advanced.h"

// compute the MC simulation
int main(int argc, char** argv) {
    double tau;	// optical depth: rho (mass absorbtion coefficient) * k (mass density of scatterers and absorbers) * l (length)
    double taumax; // optical depth for l = L (maximum length of the slab)
    double albedo = 1.0; // probability for a photon to scatter: 1.0 total scattering, 0.0 total absorbtion
    double sint, cost; 		// theta of a photon
    double sinp, cosp; 		// phi of a photon
    double phi;			// phi of a photon
    double theta;		// theta of a photon
    double xtot, ytot, ztot; 	// new position of a photon
    double x1, y1, z1; 		// photon initial position
    double x2, y2, z2;		// photon final position
    double s;			// length crossed by a photon
    int aflag;			// photon was absorbed
    double nxp,nyp,nzp;	// direction of photon in Cartesian coordinates
    int tflag; 		// photon exited the grid
    double wgt1;	//weight for forced scattering

    string imname, momname;	//the output files

    // MPI stuff
    int ret, nproc, rank, length, period = 0;
    char proc_name[MPI_MAX_PROCESSOR_NAME];
    MPI_Status status;

    double* glob_intensity;
    double* glob_energy;
    double* glob_erri;
    double* glob_sigmai;
    double* glob_hplus;
    double* glob_hminus;
    double* glob_jplus;
    double* glob_jminus;
    double* glob_kplus;
    double* glob_kminus;
    int image_pixels;
    int* glob_image;

    image_size = 100;
    density = 1.0;
    opacity = 1.0;
    distanc = 2.0;
    depth = 1.0;
    hgg = 0.41;
    pl = 0.51;
    pc = 0.0;
    sc = 1.0;
    g2 = hgg * hgg;
    xdim =20;
    ydim =20;
    zdim =20;
    xmax =1.0;
    ymax =1.0;
    zmax =1.0;
    imname="/home/eangelou/image.dat";
    momname="/home/eangelou/moments.dat";

    // load parameters from the parameters file
    string line;
    ifstream parfile;

    /* init MPI */
    MPI_Init(&argc, &argv);
    /* get task rank */
    MPI_Comm_rank(MPI_COMM_WORLD, &rank);
    /* get total number of tasks */
    MPI_Comm_size(MPI_COMM_WORLD, &nproc);
    /* get processor name */
    MPI_Get_processor_name(proc_name, &length);

    //Where am I?
    cout << "[Rank:"<<rank<<"] on "<<proc_name<<endl;

    if (rank == 0) {

        if (argc == 2) {
            parfile.open(argv[1], ifstream::in);
        } else {
            cerr << "Wrong arguments: usage: " << argv[0] << " <parameterFile>" << endl;
            exit(-1);
        }

        list<string> ls;

        if (parfile.is_open()) {
            while (! parfile.eof()) {
                getline(parfile, line);

                // if the line is not a comment and is not empty
                if(line.find("#",0) == string::npos && line.find(";", 0) != string::npos) {

                    //tokenize the line
                    char nline[150];
                    line.copy(nline, line.length());

                    char* pch = strtok(nline, " ,=;");

                    while(pch != NULL) {
                        ls.push_back(pch);
                        pch = strtok(NULL, " ,=;");
                    }

                    //Search for the variables
                    for(list<string>::const_iterator i = ls.begin();i != ls.end(); ++i) {
                        if(*i == "nphotons") {
                            ++i;
                            nphotons = atoi(((string)*i).c_str());
                        } else if(*i == "seed") {
                            ++i;
                            seed = atoi(((string)*i).c_str());
                        } else if(*i == "mubins") {
                            ++i;
                            mubins = atoi(((string)*i).c_str());
                        } else if(*i == "nlevel") {
                            ++i;
                            nlevel = atoi(((string)*i).c_str());
                        } else if(*i == "taumax") {
                            ++i;
                            taumax = atof(((string)*i).c_str());
                        } else if(*i == "albedo") {
                            ++i;
                            albedo = atof(((string)*i).c_str());
                        } else if(*i == "density") {
                            ++i;
                            density = atof(((string)*i).c_str());
                        } else if(*i == "opacity") {
                            ++i;
                            opacity = atof(((string)*i).c_str());
                        } else if(*i == "image") {
                            ++i;
                            image_size = atoi(((string)*i).c_str());
                        } else if(*i == "distance") {
                            ++i;
                            distanc = atof(((string)*i).c_str());
                        } else if(*i == "pl") {
                            ++i;
                            pl = atof(((string)*i).c_str());
                        } else if(*i == "pc") {
                            ++i;
                            pc = atof(((string)*i).c_str());
                        } else if(*i == "sc") {
                            ++i;
                            sc = atof(((string)*i).c_str());
                        } else if(*i == "hgg") {
                            ++i;
                            hgg = atof(((string)*i).c_str());
                        } else if(*i == "obs_theta") {
                            ++i;
                            obs_theta = atof(((string)*i).c_str());
                        } else if(*i == "obs_phi") {
                            ++i;
                            obs_phi = atof(((string)*i).c_str());
                        } else if(*i == "xmax") {
                            ++i;
                            xmax = atof(((string)*i).c_str());
                        } else if(*i == "ymax") {
                            ++i;
                            ymax = atof(((string)*i).c_str());
                        } else if(*i == "zmax") {
                            ++i;
                            zmax = atof(((string)*i).c_str());
                        } else if(*i == "xdim") {
                            ++i;
                            xdim = atoi(((string)*i).c_str());
                        } else if(*i == "ydim") {
                            ++i;
                            ydim = atoi(((string)*i).c_str());
                        } else if(*i == "zdim") {
                            ++i;
                            zdim = atoi(((string)*i).c_str());
                        } else if(*i == "imname") {
                            ++i;
                            imname = (string)*i;
                        } else if(*i == "momname") {
                            ++i;
                            momname = (string)*i;
                        }
                    }
                }
            }

            //close the file
            parfile.close();
        } else {
            cout << "NO PARAMETER FILE: STOPPING"<<endl;
            exit(-1);
        }

        cout << "Sending parameters..." <<endl;
    }

    /* send the boundaries to each processor */
    MPI_Bcast(&seed, 1, MPI_INT, 0, MPI_COMM_WORLD);
    MPI_Bcast(&mubins, 1, MPI_INT, 0, MPI_COMM_WORLD);
    MPI_Bcast(&nlevel, 1, MPI_INT, 0, MPI_COMM_WORLD);
    MPI_Bcast(&albedo, 1, MPI_DOUBLE, 0, MPI_COMM_WORLD);
    MPI_Bcast(&nphotons, 1, MPI_INT, 0, MPI_COMM_WORLD);
    MPI_Bcast(&density, 1, MPI_DOUBLE, 0, MPI_COMM_WORLD);
    MPI_Bcast(&opacity, 1, MPI_DOUBLE, 0, MPI_COMM_WORLD);
    MPI_Bcast(&taumax, 1, MPI_DOUBLE, 0, MPI_COMM_WORLD);
    MPI_Bcast(&pl, 1, MPI_DOUBLE, 0, MPI_COMM_WORLD);
    MPI_Bcast(&pc, 1, MPI_DOUBLE, 0, MPI_COMM_WORLD);
    MPI_Bcast(&sc, 1, MPI_DOUBLE, 0, MPI_COMM_WORLD);
    MPI_Bcast(&hgg, 1, MPI_DOUBLE, 0, MPI_COMM_WORLD);
    MPI_Bcast(&obs_theta, 1, MPI_DOUBLE, 0, MPI_COMM_WORLD);
    MPI_Bcast(&obs_phi, 1, MPI_DOUBLE, 0, MPI_COMM_WORLD);
    MPI_Bcast(&xmax, 1, MPI_DOUBLE, 0, MPI_COMM_WORLD);
    MPI_Bcast(&ymax, 1, MPI_DOUBLE, 0, MPI_COMM_WORLD);
    MPI_Bcast(&zmax, 1, MPI_DOUBLE, 0, MPI_COMM_WORLD);
    MPI_Bcast(&xdim, 1, MPI_INT, 0, MPI_COMM_WORLD);
    MPI_Bcast(&ydim, 1, MPI_INT, 0, MPI_COMM_WORLD);
    MPI_Bcast(&zdim, 1, MPI_INT, 0, MPI_COMM_WORLD);
    MPI_Bcast(&distanc, 1, MPI_DOUBLE, 0, MPI_COMM_WORLD);
    MPI_Bcast(&image_size, 1, MPI_INT, 0, MPI_COMM_WORLD);

    //     	cout << "[ " << rank << " ]: albedo = " << albedo << ", density = " << density << ", opacity = " << opacity << ", nphotons = " << nphotons << ", mubins = " << mubins << ", nlevel = " << nlevel << ", depth = " << depth << ", distance = " << distanc << ", image = " << image_size<< ", xdim = " << xdim<< ", ydim = " << ydim<< ", zdim = " << zdim << endl;

    //distribute photons to processes more or less fairly
    if (rank<nphotons%nproc) {
        nphotons /= nproc;
        nphotons++;
    } else {
        nphotons /= nproc;
    }

    srandom(seed  + 5 * rank);

    // initialize arrays
    init_array();

    int last=0;

    for (int i = 0; i < nphotons; i++) {

        //Print messages
        if (i%10000 == 0 && rank==0) {
            if (i!=last) {
                cout << i*nproc <<" stellar photons completed"<<endl;
                last=i;
            }
        }

        // throw a new photon isotropically from the origin
        newphoton(sint, cost, sinp, cosp, theta, phi, xtot, ytot, ztot, nxp, nyp, nzp, xcell, ycell, zcell);

        if(DEBUG)
            cout << "[ " << rank << " ] Photon: theta = " << acos(cost) * 180 / M_PI << ", phi = " <<  phi * 180 / M_PI << "..." << flush;

        // absorb / scattering flag
        aflag = 0;

        //         double tau1;

        //find maximum optical depth the photon can travel
        findTauMax( xtot,  ytot,  ztot,  nxp,  nyp,  nzp,  xmax,  ymax,  zmax,  xfaces,  yfaces,  zfaces,  taumax);

        if (taumax < .001) {
            i--;
            continue;
        }

        double wgt1 = 1.0 - exp(-taumax);

        /* ******* Force photon to scatter at optical depth tau before edge of grid once */
        s = -log(1.0 - urand() * wgt1);

        tflag = 0;
        int scat =0;
        // while the photon is inside the slab
        while (tflag==0 && scat<250) { //(ztot >= 0 && ztot <= depth) {
            scat++;
            x1 = xtot;
            y1 = ytot;
            z1 = ztot;

            // sample a traveled distance s (only for subsequent scattering)
            if (scat>1) {
                // s = - log(urand()) / (density * opacity);
                s = -log(urand());
            }


            // check if scatter or absorb
            if(urand() < albedo) {	// scattering
                // 	  double pi = M_PI;
                // 	  double twopi = 2 * M_PI;
                //cout << "PRESCATT: Phi = " << (phi * 180/M_PI) << ", Theta = " << (acos(cost) * 180/M_PI) << endl;
                anisoscat(xtot, ytot, ztot, sint, cost, sinp, cosp, phi, fi, fq, fu, fv, pl, pc, sc, hgg, g2);

                // 	  isoscatt(sint, cost, sinp, cosp, theta, phi, nxp, nyp, nzp);
                // 	  cout << "help!!!" << endl;
                //cout << "POSTSCATT: Phi = " << (phi * 180/M_PI) << ", Theta = " << (acos(cost) * 180/M_PI) << endl << endl;

                integration(xtot,  ytot,  ztot,  nxp,  nyp,  nzp,  xmax,  ymax,  zmax,  xfaces,  yfaces,  zfaces,  tflag);

                x2 = xtot;
                y2 = ytot;
                z2 = ztot;

                // recompute the moments
                moments(x1, y1, z1, x2, y2, z2, cost, nlevel);


                if(DEBUG)
                    cout << "S" << flush;
            } else {			// absorbtion
                aflag=1;
                if(DEBUG)
                    cout << "A" << endl;
                break;
            }
        }

        // test whether photon is "inside star", if it is start a new photon
        if(ztot < 0) {
            --i;
            if(DEBUG)
                cout << "_BACK" << endl;
            continue;
        }

        // the photon is exiting from behind the slab and it will be visible to the observer
        if(aflag == 0) {
            int j = (int)(mubins * (cost + 1.0)/2.0) + 1;
            erri[j] = erri[j] + 1.0;
            energy[j] = energy[j] + 1.0;
            if(DEBUG)
                cout << "_OUT" << flush;
            double x, y;
            double ratio = (2.0 * distanc) / image_size;
            int ximage, yimage;

            // 	double rho = sqrt(xtot * xtot + ytot * ytot + ztot * ztot);
            // 	x = rho * sinp * cost;
            // 	y = rho * sinp * sint;

            // 	double transk = ztot / depth;

            // 	xtot *= transk;
            // 	ytot *= transk;
            // 	ztot = depth;

            x = ztot * obs_sint - ytot * obs_cost * obs_sinp - xtot * obs_cost * obs_cosp;
            y = ytot * obs_cosp - xtot * obs_sinp;

            //cout << xtot << " " << ytot << "\n";
            //cout << x << " " << y << "\n";

            ximage = (int)((x + distanc) / ratio);
            yimage = (int)((y + distanc) / ratio);

            if(ximage >= 0 && ximage < image_size && yimage >= 0 && yimage < image_size)
                image[ximage * image_size + yimage]++;

            //cout << "Photon exiting with theta = " << acos(1 - ((double)l/mubins)) * 180/M_PI << ", phi = " << phi * 180 / M_PI << endl;
        }
    }

    // for(int i = 0; i < image_size; ++i) {
    //             for(int j = 0; j < image_size; ++j) {

    //                 cout <<"[Rank:"<<rank<<"] "<< i << " " << j << " " << image[i * image_size + j] << endl;
    //             }
    //             //cout << endl;
    //         }
    image_pixels = image_size * image_size;

    // if (rank==0){
    // cout <<"[Rank:"<<rank << "]  pixels="<<image_pixels<<endl;
    glob_image = new int[image_pixels];
    for (int p=0; p<image_pixels; p++) {
        glob_image[p] =0;
    }
    // }

    glob_intensity = new double[mubins];
    glob_energy = new double[mubins];
    glob_erri = new double[mubins];
    glob_sigmai = new double[mubins];


    glob_hminus = new double[nlevel];
    glob_hplus = new double[nlevel];
    glob_jplus = new double[nlevel];
    glob_jminus = new double[nlevel];
    glob_kplus = new double[nlevel];
    glob_kminus = new double[nlevel];


    //wait for everyone to finish
    if (rank == 0)
        cout<< "Synching and gathering results..."<<endl;
    MPI_Barrier(MPI_COMM_WORLD);


    /* get the final result */
    MPI_Reduce(image, glob_image, image_pixels, MPI_INT, MPI_SUM, 0, MPI_COMM_WORLD);
    MPI_Reduce(intensity, glob_intensity, mubins, MPI_DOUBLE, MPI_SUM, 0, MPI_COMM_WORLD);
    MPI_Reduce(energy, glob_energy, mubins, MPI_DOUBLE, MPI_SUM, 0, MPI_COMM_WORLD);
    MPI_Reduce(erri, glob_erri, mubins, MPI_DOUBLE, MPI_SUM, 0, MPI_COMM_WORLD);
    MPI_Reduce(sigmai, glob_sigmai, mubins, MPI_DOUBLE, MPI_SUM, 0, MPI_COMM_WORLD);
    MPI_Reduce(hplus, glob_hplus, nlevel, MPI_DOUBLE, MPI_SUM, 0, MPI_COMM_WORLD);
    MPI_Reduce(hminus, glob_hminus, nlevel, MPI_DOUBLE, MPI_SUM, 0, MPI_COMM_WORLD);
    MPI_Reduce(jplus, glob_jplus, nlevel, MPI_DOUBLE, MPI_SUM, 0, MPI_COMM_WORLD);
    MPI_Reduce(jminus, glob_jminus, nlevel, MPI_DOUBLE, MPI_SUM, 0, MPI_COMM_WORLD);
    MPI_Reduce(kplus, glob_kplus, nlevel, MPI_DOUBLE, MPI_SUM, 0, MPI_COMM_WORLD);
    MPI_Reduce(kminus, glob_kminus, nlevel, MPI_DOUBLE, MPI_SUM, 0, MPI_COMM_WORLD);


    // output the results
    if (rank == 0) {
        cout <<"Writing results..."<<endl;
        fstream file_op(imname.c_str(),ios::out);
        int max_intensity = 0;
        for(int i = 0; i < image_size; ++i) {
            for(int j = 0; j < image_size; ++j) {
                //if(image[i][j] > max_intensity)
                //max_intensity = image[i][j];
                //image[i * image_size + j];
                file_op << i << " " << j << " " << glob_image[i * image_size + j] << endl;
                //cout << image[i * image_size + j] << " ";
            }
            //cout << endl;
        }
        file_op << flush;
        file_op.close();

        memcpy(intensity, glob_intensity, sizeof(double) * mubins);
        memcpy(energy, glob_energy, sizeof(double) * mubins);
        memcpy(erri, glob_erri, sizeof(double) * mubins);
        memcpy(sigmai, glob_sigmai, sizeof(double) * mubins);
        memcpy(hminus, glob_hminus, sizeof(double) * nlevel);
        memcpy(hplus, glob_hplus, sizeof(double) * nlevel);
        memcpy(jminus, glob_jminus, sizeof(double) * nlevel);
        memcpy(jplus, glob_jplus, sizeof(double) * nlevel);
        memcpy(kminus, glob_kminus, sizeof(double) * nlevel);
        memcpy(kplus, glob_kplus, sizeof(double) * nlevel);

	nphotons *= nproc;
        output(momname);

        cout <<"Shutting down..."<<endl;
    }



    // cout <<"[Rank:"<<rank << "]  OK!!!!"<<endl;
    MPI_Barrier(MPI_COMM_WORLD);

    // Buggy MPI????
    MPI_Finalize();

    /* exit succesfully */
    exit(0);
}


